/******************************************************************************

Project:
    DirectNet MPF

File:
    devDnMpf.c

Description:
    Common device support routines for directNet over MPF

Author:
    Andrew Johnson
Created:
    September 2001
Version:
    devDnMpf.c,v 1.1 2002/01/07 22:36:17 anj Exp

******************************************************************************/

#include <vxWorks.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <recSup.h>
#include <devSup.h>
#include <drvSup.h>
#include <dbScan.h>
#include <menuScan.h>

#include "directNetClient.h"
#include "devDnMpf.h"


/* Create a drvet for the report function */
static long report();

struct drvet drvDnMpf = {2, report, NULL};


/* Global variables */

int devDnMpfDebug = 0;


/* Module variables */

static struct plcInfo *dnMpf_plcs;


/* Find a named PLC */

struct plcInfo *dnMpfPlc(const char* pname) {
    struct plcInfo *pPlc = dnMpf_plcs;
    
    while (pPlc) {
	if (strcmp(pname, pPlc->name) == 0)
	    return pPlc;
	pPlc = pPlc->pNext;
    }
    return NULL;
}


/* Parse an instio link field */

int dnMpfAddr(struct dbCommon *prec, struct plcAddr *paddr, struct link *plink) {
    struct instio *pinstio;
    struct plcInfo *pPlc;
    unsigned int addr;
    char *parse;
    int addrType = -1;
    int i;
    
    const static struct {
    	const char *letters;	/* How do we introduce it? */
	enum {BIT, WORD} aType;	/* Whether addr refers to bits or words */
	unsigned int offset;	/* Starting V-memory address (NB octal) */
	unsigned int maxAddr;	/* largest legal bit/address */
    } aTypes[] = {	/* NB: Relative order of entries is important */
    	{"B",   WORD, 000000, 041237},
	{"CTA", WORD, 001000,    127},
	{"CT",  BIT,  041140,    127},
	{"C",   BIT,  040600,   1023},
	{"SP",  BIT,  041200,    511},
	{"S",   BIT,  041000,   1023},
	{"TA",  WORD, 000000,    255},
	{"T",   BIT,  041100,    255},
    	{"V",   WORD, 000000, 041237},
	{"X",   BIT,  040400,    511},
	{"Y",   BIT,  040500,    511},
    };
    
    /* link address must be INST_IO */
    if (plink->type != INST_IO) {
    	recGblRecordError(S_dev_badBus, (void *) prec,
    	    "devDnMpf (init_record) Illegal Bus Type");
    	return S_dev_badBus;
    }
    
    /* check the string starts with a recognized PLC name */
    pinstio = &plink->value.instio;
    parse = pinstio->string;
    for (pPlc = dnMpf_plcs; pPlc; pPlc = pPlc->pNext) {
	i = strlen(pPlc->name);
	if ((strncmp(parse, pPlc->name, i) == 0) &&
	    parse[i] == ' ') {
	    parse += i + 1;
	    break;
	}
    }
    if ((pPlc == 0) || (parse == pinstio->string)) {
	recGblRecordError(S_dev_badCard, (void *) prec,
	    "devDnMpf (init_record) named PLC not found");
	return S_dev_badCard;
    }
    paddr->plcInfo = pPlc;
    
    /* parse the PLC address type */
    for (i=0; i < sizeof(aTypes) / sizeof(aTypes[0]); i++) {
    	if (strncmp(parse, aTypes[i].letters, strlen(aTypes[i].letters)) == 0) {
	    addrType = i;
	    break;
	}
    }
    if (addrType < 0) {
    	recGblRecordError(S_dev_badSignal, (void *) prec,
    	    "devDnMpf (init_record) PLC address type not recognized");
	return S_dev_badSignal;
    }
    parse += strlen(aTypes[addrType].letters);
    
    /* digits giving the PLC address */
    addr = strtoul(parse, &parse, 8);
    if (addr > aTypes[addrType].maxAddr) {
    	recGblRecordError(S_dev_badSignal, (void *) prec,
		"devDnMpf (init_record) PLC address too large");
	return S_dev_badSignal;
    }
    
    if (aTypes[addrType].aType == BIT) {
    	paddr->vAddr = aTypes[addrType].offset + (addr / PLCWORDBITS) + DNREFOFFSET;
	paddr->bitNum = addr % PLCWORDBITS;
	/* This code is wrong if offset is not a multiple of 16 */
    } else {
    	paddr->vAddr = addr + aTypes[addrType].offset + DNREFOFFSET;
	
	if (*parse == '.') paddr->bitNum = strtoul(++parse, &parse, 8);
	else paddr->bitNum = 0;
    }
    
    return 0;
}


/* Create plc information entry */

int createDnMpfPLC(const char* pname, int slaveId, const char* pserver, int location) {
    struct plcInfo *pPlc = dnMpf_plcs;
    
    if ((slaveId <= 0) || (slaveId > MAXDNSLAVEID)) {
	printf("createDnMpfPLC: Slave ID out of range 1 .. %d\n", MAXDNSLAVEID);
	return ERROR;
    }
    
    /* Check for duplicates */
    while (pPlc) {
	if (strcmp(pname, pPlc->name) == 0) {
	    printf("createDnMpfPLC: Duplicate name to PLC at \"%s\" ID %u\n",
		   pPlc->server, pPlc->slaveId);
	    return ERROR;
	}
	if ((slaveId == pPlc->slaveId) &&
	    (strcmp(pserver, pPlc->server) == 0)) {
	    printf("createDnMpfPLC: Duplicate address to PLC \"%s\"\n",
		   pPlc->name);
	    return ERROR;
	}
	pPlc = pPlc->pNext;
    }
    
    /* Create and populate a new info structure for this plc */
    pPlc = (struct plcInfo *) (malloc(sizeof(struct plcInfo)));
    if (pPlc == NULL) {
	perror("createDnMpfPLC");
	return ERROR;
    }

    pPlc->name     = pname;
    pPlc->server   = pserver;
    pPlc->location = location;
    pPlc->slaveId  = slaveId;
    pPlc->alarm    = NO_ALARM;
    pPlc->rdCache  = NULL;
    pPlc->wrCache  = NULL;
    pPlc->nRdReqs  = 0;
    pPlc->nWrReqs  = 0;
    pPlc->nSuccess = 0;
    pPlc->nDnFail  = 0;
    pPlc->nMpFail  = 0;

    /* Add it to the list */
    pPlc->pNext    = dnMpf_plcs;
    dnMpf_plcs = pPlc;
    
    return OK;
}


/* Report functions */

void dnMpfReport(int detail, dnPlcReportFn ioReport) {
    struct plcInfo *pPlc = dnMpf_plcs;
    
    while (pPlc) {
	printf("PLC \"%s\" via MPF server \"%s\" (%d) with DirectNet ID %u\n",
    		pPlc->name, pPlc->server, pPlc->location, pPlc->slaveId);
    	switch (detail) {
	    case 0:
    	    	break;
		
	    case 1:
    	    	printf("    alarm = %hu, nRdReqs = %lu, nWrReqs = %lu\n",
    			pPlc->alarm, pPlc->nRdReqs, pPlc->nWrReqs);
    	    	printf("    nSuccess = %lu, nDnFail = %lu, nMpFail = %lu\n",
    			pPlc->nSuccess, pPlc->nDnFail, pPlc->nMpFail);
    	    	break;
		
	    default:
	    	if (ioReport != NULL)
	    	    (*ioReport)(detail, pPlc);
	}
	
    	pPlc = pPlc->pNext;
    }
}

static long report(int detail) {
    if (detail <= 1)
    	dnMpfReport(detail, NULL);
    return 0;
}
