/******************************************************************************

Project:
    DirectNet MPF

File:
    devXiDnMpf.c

Description:
    Input device support routines for directNet over MPF

Author:
    Andrew Johnson
Created:
    August 2001
Version:
    devXiDnMpf.c,v 1.1 2002/01/07 22:36:18 anj Exp

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <taskLib.h>
#include <tickLib.h>

#include <recSup.h>
#include <devSup.h>
#include <dbScan.h>
#include <drvSup.h>
#include <errlog.h>
#include <aiRecord.h>
#include <biRecord.h>
#include <mbbiRecord.h>
#include <mbbiDirectRecord.h>
#include <menuScan.h>

#include "directNetClient.h"
#include "directNetMpf.h"
#include "devDnMpf.h"


static long report();
static long init_ai();
static long init_bi();
static long init_mbbi();
static long init_mbbid();
static long get_ioint();
static long read_data();

/* XXDSET devXxDnMpf= { 5,report,init,init_rec, get_ioint,read_data}; */
XXDSET devAiDnMpf    = { 6,NULL,  NULL,init_ai,   get_ioint,read_data, NULL };
XXDSET devBiDnMpf    = { 5,report,NULL,init_bi,   get_ioint,read_data };
XXDSET devMbbiDnMpf  = { 5,NULL,  NULL,init_mbbi, get_ioint,read_data };
XXDSET devMbbidDnMpf = { 5,NULL,  NULL,init_mbbid,get_ioint,read_data };


struct rdCache {
    struct rdCache *pNext;
    struct rdItem {
	struct plcMessage msg;	/* This *MUST* be the first item */
	unsigned char rdData[DN_RDDATA_MAX];
	unsigned int startAddr;
	unsigned char nWords;
	unsigned char active;
	unsigned short data[DN_RDDATA_MAX / DN_PLCWORDLEN];
	unsigned long timestamp;
	IOSCANPVT intInfo;
	struct dpvtIn *recList;
    } item;
};

struct dpvtIn {
    struct dbCommon *precord;
    enum recType {AI, BI, MBBI, MBBID} type;
    unsigned char waiting;
    struct plcAddr plcAddr;
    struct plcInfo *plcInfo;
    struct rdItem *rdItem;
    struct dpvtIn *recNext;
    unsigned long tickStart;
};

static const char *recTypeName[] = {
    "ai", "bi", "mbbi", "mbbiDirect"
};

static void ioReport(int detail, struct plcInfo *pPlc) {
    struct rdCache *pcache = pPlc->rdCache;
    
    while (pcache) {
	switch (detail) {
	case 2:
	    printf("    RdCache for V%o - V%o last updated at tick %ld\n",
		   pcache->item.startAddr - DNREFOFFSET,
		   pcache->item.startAddr - DNREFOFFSET + pcache->item.nWords - 1,
		   pcache->item.timestamp);
	    break;
	    
	case 3: {
		int i;
	
		printf("    RdCache buffer for V%o - V%o holds (hex):\n       ",
			pcache->item.startAddr - DNREFOFFSET,
			pcache->item.startAddr - DNREFOFFSET + pcache->item.nWords - 1);
		for (i=0; i < pcache->item.nWords; i++) {
		    printf(" %04x", pcache->item.data[i]);
		}
		putchar('\n');
	    } 
	    break;
	    
	case 4: {
		struct dpvtIn *dpvt = pcache->item.recList;
	
		printf("    Used by:\n");
		while (dpvt) {
		    printf("	    %s (%s)",
			    dpvt->precord->name, recTypeName[dpvt->type]);
		    if (dpvt->plcAddr.bitNum)
			printf(" at B%o.%o\n",
			       dpvt->plcAddr.vAddr - DNREFOFFSET,
			       dpvt->plcAddr.bitNum);
		    else
			printf(" at V%o\n",
			       dpvt->plcAddr.vAddr - DNREFOFFSET);
	
		    dpvt = dpvt->recNext;
		}
	    }
	    break;
	}
	pcache = pcache->pNext;
    }
    
    return;
}

static long report(int detail) {
    if (detail > 1) dnMpfReport(detail, ioReport);
    return 0;
}


static void devXiDnCallback(struct plcMessage *pMsg) {
    /* This uses a kludge, we actually need the address of the struct rdItem.
       The two must be identical or this will fail. */
    struct rdItem *pitem = (struct rdItem *) pMsg;
    struct dpvtIn *dpvt = pitem->recList;
    struct plcInfo *pPlc = dpvt->plcInfo;
    int i;
    
    if (devDnMpfDebug >= 10) {
	printf("devXiDnMpf: Got a reply from PLC \"%s\"\n", 
		pPlc->name);
	printf("devXiDnMpf: Read V%o[%d] returned status %d\n",
		pMsg->addr, pMsg->len, pMsg->status);
    } else if (pMsg->status && devDnMpfDebug >= 5)
	printf("devXiDnMpf: Read reply from PLC \"%s\" has status %d\n",
		pPlc->name, pMsg->status);
    
    switch (pMsg->status) {
    case DN_SUCCESS:
	/* Read succeeded */
	pPlc->alarm = NO_ALARM;
	pPlc->nSuccess++;
	break;
	
    case DN_RETRY:
    case DN_TIMEOUT:
    case DN_RETRIED:
    case DN_BADARGS:
    case DN_ABORT:
	/* DirectNet I/O failed */
	errlogPrintf("devXiDnMpf: Comms error with PLC \"%s\" status %d\n",
			pPlc->name, pMsg->status);
	pPlc->alarm = INVALID_ALARM;
	pPlc->nDnFail++;
	break;
	
    default:	/* MPF or other error occurred */
	errlogPrintf("devXiDnMpf: Error from MPF server \"%s\" status %d\n",
			pMsg->server, pMsg->status);
	pPlc->alarm = MAJOR_ALARM;
	pPlc->nMpFail++;
    }
    
    if (pPlc->alarm == NO_ALARM) {
	if (devDnMpfDebug >= 10) {
	    printf("devXiDnMpf: Got read reply: cmd=%#x addr=%#o, len=%d, status=%d",
		   pMsg->cmd, pMsg->addr, pMsg->len, pMsg->status);
	    for (i=0; i < pMsg->len; i++) {
		printf("\ndevXiDnMpf: read data[%d] = ", i);
		while ((i < pMsg->len) && ((i+1) & 7))
		    printf ("%2.2x ", pMsg->pdata[i++]);
	    }
	    printf("\n");
	}

	/* Cache the reply data */
	for (i=0; i < pitem->nWords; i++) {
	    pitem->data[i] = (pMsg->pdata[2*i+1] << 8) |
			      pMsg->pdata[2*i];
	}
    }
    pitem->timestamp = tickGet();
    pitem->active = FALSE;
    
    /* Now process all the waiting records */
    while (dpvt != NULL) {
	struct dbCommon *prec = dpvt->precord;
	struct rset *prset = prec->rset;
	if (devDnMpfDebug >= 15) {
	    printf("Examining \"%s\", waiting = %d\n", prec->name, dpvt->waiting);
	}
	if (dpvt->waiting) {
	    dpvt->waiting = FALSE;
	    
	    dbScanLock(dpvt->precord);
	    (*prset->process)(dpvt->precord);
	    dbScanUnlock(dpvt->precord);
	}
	dpvt = dpvt->recNext;
    }
    
    /* Finally trigger any I/O Interrupt records */
    scanIoRequest(pitem->intInfo);
}


static long init_input(struct dbCommon *prec, enum recType type, struct link *plink) {
    struct dpvtIn *dpvt;
    struct rdCache **ppcache, *pcache;
    struct plcInfo *pPlc;
    unsigned int addr;
    long status;
    const int maxWords = DN_RDDATA_MAX / DN_PLCWORDLEN;
    
    if (devDnMpfDebug > 0)
	printf ("devXiDnMpf: init_input invoked for \"%s\"\n", prec->name);
    
    dpvt = (struct dpvtIn *) (calloc(1, sizeof(struct dpvtIn)));
    if (dpvt == NULL) {
	errlogPrintf("devXiDnMpf: calloc failed for \"%s\"\n", prec->name);
	prec->pact = TRUE;
	return S_rec_outMem;
    }
    prec->dpvt = (void *) dpvt;
    
    dpvt->precord = prec;
    dpvt->type    = type;
    dpvt->waiting = FALSE;
    
    status = dnMpfAddr(prec, &dpvt->plcAddr, plink);
    if (status) {
	prec->pact = TRUE;
	return status;
    }
    pPlc = dpvt->plcAddr.plcInfo;
    addr = dpvt->plcAddr.vAddr;
    
    dpvt->plcInfo = pPlc;
    
    if (devDnMpfDebug > 10)
	printf ("devXiDnMpf: dpvt = %p, vAddr = V%o\n", 
		dpvt, addr - DNREFOFFSET);

    /* Search for an existing cache entry or one that can be extended */
    ppcache = &(pPlc->rdCache);
    pcache = NULL;
    while (*ppcache) {
	if ((addr >= (*ppcache)->item.startAddr + (*ppcache)->item.nWords - maxWords) &&
	    ((*ppcache)->item.startAddr + maxWords > addr)) {
	    if (devDnMpfDebug > 10)
		printf ("devXiDnMpf: Suitable rdCache entry found: V%o\n", 
			(*ppcache)->item.startAddr - DNREFOFFSET);
	    pcache = *ppcache;
	    break;
	}
	ppcache = &(*ppcache)->pNext;
    }
    
    if (pcache == NULL) {
	struct plcMessage *pMsg;
	
	/* Not found, create a new entry */
	pcache = (struct rdCache *) calloc(1, sizeof (struct rdCache));
	if (pcache == NULL) {
	    errlogPrintf("devXiDnMpf: calloc failed for \"%s\"\n", prec->name);
	    prec->pact = TRUE;
	    return S_rec_outMem;
	    /* Really ought to free dpvt first... */
	}
	
	if (devDnMpfDebug > 10)
	    printf ("devXiDnMpf: new rdCache entry created for V%o\n", 
		    addr - DNREFOFFSET);
	
	/* Initialise: cache entry */
	pcache->item.startAddr = addr;
	pcache->item.nWords = 1;
	pcache->item.active = FALSE;
	pcache->item.timestamp = 0;
	scanIoInit(&pcache->item.intInfo);
	
	/* plcMessage entry */
	pMsg = &pcache->item.msg;
	pMsg->server   = pPlc->server;
	pMsg->location = pPlc->location;
	pMsg->callback = devXiDnCallback;
	pMsg->cmd      = (pPlc->slaveId << 8) | READVMEM;
	pMsg->len      = DN_PLCWORDLEN;
	pMsg->addr     = addr;
	pMsg->pdata    = pcache->item.rdData;
	
	/* creat a new MPF client object */
	status = initDnMpfClient(pMsg);
	if (status) {
	    prec->pact = TRUE;
	    return status;
	}
	
	/* Finally install the various list links */
	dpvt->recNext = NULL;
	pcache->pNext = NULL;
	pcache->item.recList = dpvt;
	(*ppcache) = pcache;
	
    } else if (addr < pcache->item.startAddr) {
	/* Need to extend the existing entry backwards */
	if (devDnMpfDebug > 10)
	    printf ("devXiDnMpf: Extending entry back to V%o\n", 
		    addr - DNREFOFFSET);
	
	/* Adjust entry start address and length */
	pcache->item.nWords += pcache->item.startAddr - addr;
	pcache->item.startAddr = addr;
	
	/* Repeat that change in the plcMessage */
	pcache->item.msg.addr = dpvt->plcAddr.vAddr;
	pcache->item.msg.len  = pcache->item.nWords * PLCWORDBYTES;
	
	/* Finally insert this record in item's record list */
	dpvt->recNext = pcache->item.recList;
	pcache->item.recList = dpvt;
	
    } else if (addr >= pcache->item.startAddr + pcache->item.nWords) {
	/* Need to extend the existing entry forwards */
	if (devDnMpfDebug > 10)
	    printf ("devXiDnMpf: Extending entry forward to V%o\n", 
		    addr - DNREFOFFSET);
	
	/* Adjust entry length, in Bitbus message too */
	pcache->item.nWords  = addr - pcache->item.startAddr + 1;
	pcache->item.msg.len = pcache->item.nWords * PLCWORDBYTES;
	
	/* Finally insert this record in item's record list */
	dpvt->recNext = pcache->item.recList;
	pcache->item.recList = dpvt;
	
    } else {
	if (devDnMpfDebug > 10)
	    printf ("devXiDnMpf: No change needed for V%o\n", 
		    addr - DNREFOFFSET);
	
	/* Just need to insert this record in item's record list */
	dpvt->recNext = pcache->item.recList;
	pcache->item.recList = dpvt;
    }
    
    /* Make sure we can get to the cache line from the record */
    dpvt->rdItem = &pcache->item;
    return 0;
}


static long init_ai(struct aiRecord *prec) {
    long status;

    if (devDnMpfDebug > 0)
	printf ("devXiDnMpf: Init ai called for \"%s\"\n", prec->name);

    status = init_input((struct dbCommon *) prec, AI, &prec->inp);
    if (status)
	return status;
    
    /* set linear conversion slope & offset, 12 bits assumed (WRONG?) */
    prec->eslo = (prec->eguf - prec->egul)/4095.0;

    return 0;
}


static long init_bi(struct biRecord *prec) {
    long status;
    
    if (devDnMpfDebug > 0)
       printf ("devXiDnMpf: Init bi called for \"%s\"\n", prec->name);
 
    status = init_input((struct dbCommon *) prec, BI, &prec->inp);
    if (status)
	return status;
    
    prec->mask = 1 << ((struct dpvtIn *)prec->dpvt)->plcAddr.bitNum;
    return 0;
}


static long init_mbbi(struct mbbiRecord *prec) {
    long status;

    if (devDnMpfDebug > 0)
	printf ("devXiDnMpf: Init mbbi called for \"%s\"\n", prec->name);
    
    status = init_input((struct dbCommon *) prec, MBBI, &prec->inp);
    if (status)
	return status;
    
    prec->shft = ((struct dpvtIn *)prec->dpvt)->plcAddr.bitNum;
    prec->mask <<= prec->shft;
    return 0;
}


static long init_mbbid(struct mbbiDirectRecord *prec) {
    long status;

    if (devDnMpfDebug > 0)
	printf ("devXiDnMpf: Init mbbiDirect called for \"%s\"\n", prec->name);
    
    status = init_input((struct dbCommon *) prec, MBBID, &prec->inp);
    if (status)
	return status;
    
    prec->shft = ((struct dpvtIn *)prec->dpvt)->plcAddr.bitNum;
    prec->mask <<= prec->shft;
    return 0;
}


static long get_ioint(int cmd, struct dbCommon *prec, IOSCANPVT *ppvt) {
    struct dpvtIn *dpvt=(struct dpvtIn *)prec->dpvt;
    struct rdItem *pitem = dpvt->rdItem;
    
    if (devDnMpfDebug >= 35)
	printf ("devXiDnMpf: get_ioint called for \"%s\"\n", prec->name);
    
    *ppvt = pitem->intInfo;
    return 0;
}



static void get_data(struct dbCommon *prec) {
    struct dpvtIn *dpvt=(struct dpvtIn *)prec->dpvt;
    struct rdItem *pitem = dpvt->rdItem;
    unsigned long value;
    
    if (devDnMpfDebug >= 35)
	printf ("devXiDnMpf: get_data called for \"%s\"\n", prec->name);
    
    /* Deal with alarms first */
    if (dpvt->plcInfo->alarm) {
	if (devDnMpfDebug >= 35)
	    printf ("devXiDnMpf: alarm status %d\n", dpvt->plcInfo->alarm);
	recGblSetSevr(prec, READ_ALARM, dpvt->plcInfo->alarm);
	return;
    }
    
    /* Find our particular number */
    value = pitem->data[dpvt->plcAddr.vAddr - pitem->startAddr];

    if (devDnMpfDebug >= 35)
	printf ("devXiDnMpf: Raw value = %#lx\n", value);

    switch(dpvt->type) {
	struct biRecord *bi;
	struct aiRecord *ai;
	struct mbbiRecord *mbbi;
	struct mbbiDirectRecord *mbbid;

	case AI:
	    ai=(struct aiRecord *)prec;
	    ai->rval = value;
	    break;

	case BI:
	    bi = (struct biRecord *)prec;
	    bi->rval = value & bi->mask;
	    break;

	case MBBI:
	    mbbi = (struct mbbiRecord *)prec;
	    mbbi->rval = value & mbbi->mask;
	    break;

	case MBBID:
	    mbbid = (struct mbbiDirectRecord *)prec;
	    mbbid->rval = value & mbbid->mask;
	    break;
    }
    return;
}


static long read_data(struct dbCommon *prec) {
    struct dpvtIn *dpvt = (struct dpvtIn *) prec->dpvt;
    struct rdItem *pitem;
    struct plcInfo *pPlc;

    if (devDnMpfDebug >= 35)
       printf ("devXiDnMpf: read_data called for \"%s\"\n", prec->name);

    if (!dpvt) return S_dev_NoInit;
    
    pitem = dpvt->rdItem;
    pPlc = dpvt->plcInfo;
    
    if (!prec->pact) {
	/* Read request.  Check the cache time */
	const static int scanRates[menuScan_1_second + 1] = { /* Units = 0.1s */
	    1, 0, 100, /* Passive, Event, I/O-Intr */
	    100, 50, 20, 10, 5, 2, 1 /* 10, 5, 2, 1, .5, .2, .1s */
	}; /* This is horrible, but there's no other easy way */
	int scanRate = 1, scanIndex = prec->scan;
	if (scanIndex <= menuScan_1_second)
	    scanRate = (scanRates[scanIndex] * vxTicksPerSecond) / 10;
	
	/* If timeNow is earlier than when the cached data gets stale
	 * (which depends on _our_ scanrate */
	if (tickGet() <= pitem->timestamp + scanRate/2) {
	    /* .. then we can use the cached data */
	    if (devDnMpfDebug >= 3)
		printf("devXiDnMpf: Using value from read cache\n");
	    
	    get_data(prec);
	    return 0;
	}
	
	/* Can't use cache data, start a Bitbus transaction */
	dpvt->tickStart = tickGet();
	dpvt->waiting = TRUE;
	
	/* Send the request */
	if (!pitem->active) {
	    if (dnMpfClientSend(&pitem->msg)) {
		recGblSetSevr(prec, WRITE_ALARM, MAJOR_ALARM);
		errlogPrintf("devXiDnMpf: MPF Send by \"%s\" failed\n", prec->name);
		pPlc->nMpFail++;
		return ERROR;
	    }
	    pitem->active = TRUE;
	    pPlc->nRdReqs++;
	}
	
	if (devDnMpfDebug >= 10)
	    printf("devXiDnMpf: Read requested at V%o\n", 
		   pitem->startAddr - DNREFOFFSET);
	
	prec->pact = TRUE;
    } else {
	/* An MPF request has completed */
	if (devDnMpfDebug >= 5)
	    printf("devXiDnMpf: alarm = %d\n", pPlc->alarm);
	
	get_data(prec);
	
	if (devDnMpfDebug >= 3)
	    printf("devXiDnMpf: Record processing for \"%s\" took %lu ticks\n",
		   prec->name, tickGet() - dpvt->tickStart);
    }

    return 0;
}
