/******************************************************************************

Project:
    DirectNet MPF

File:
    devXoDnMpf.c

Description:
    Output device support routines for directNet over MPF

Author:
    Andrew Johnson
Created:
    August 2001
Version:
    devXoDnMpf.c,v 1.1 2002/01/07 22:36:18 anj Exp

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <taskLib.h>
#include <tickLib.h>

#include <recSup.h>
#include <devSup.h>
#include <drvSup.h>
#include <errlog.h>
#include <aoRecord.h>
#include <boRecord.h>
#include <mbboRecord.h>
#include <mbboDirectRecord.h>

#include "directNetClient.h"
#include "directNetMpf.h"
#include "devDnMpf.h"


static long report();
static long init_ao();
static long init_bo();
static long init_mbbo();
static long init_mbbod();
static long write_data(struct dbCommon *prec);

/* XXDSET devXxDnMpf={ 5,report,init,init_rec,get_ioint,write_data}; */
XXDSET devAoDnMpf    = { 6,NULL,  NULL,init_ao,    NULL, write_data, NULL };
XXDSET devBoDnMpf    = { 5,report,NULL,init_bo,    NULL, write_data };
XXDSET devMbboDnMpf  = { 5,NULL,  NULL,init_mbbo,  NULL, write_data };
XXDSET devMbbodDnMpf = { 5,NULL,  NULL,init_mbbod, NULL, write_data };


struct wrCache {
    struct wrItem {
	unsigned short word;
    } item[WRITEMAXADDR - WRITEMINADDR + 1];
};

struct dpvtOut {
    struct plcMessage msg;	/* This *MUST* be the first item */
    unsigned char wrData[DN_PLCWORDLEN];
    struct dbCommon *precord;
    enum recType {AO, BO, MBBO, MBBOD} type;
    struct plcAddr plcAddr;
    struct wrItem *wrItem;
    unsigned long tickStart;
};


static void ioReport(int detail, struct plcInfo *pPlc) {
    struct wrCache *pcache = pPlc->wrCache;
    
    if (pcache) {
	printf("    WrCache for V%o - V%o starts at %p\n",
		WRITEMINADDR,
		WRITEMAXADDR,
		&(pcache->item[0].word));
    }
    return;
}

static long report(int detail) {
    if (detail > 1) dnMpfReport(detail, ioReport);
    return 0;
}

static void devXoDnCallback(struct plcMessage *pMsg) {
    struct dpvtOut  *dpvt = (struct dpvtOut *) pMsg;
    struct dbCommon *precord;
    struct rset     *prset;

    precord=dpvt->precord;
    prset=(struct rset *)(precord->rset);

    dbScanLock(precord);
    (*prset->process)(precord);
    dbScanUnlock(precord);
}


/* Record init routines */
static long init_output(struct dbCommon *prec, enum recType type, struct link *plink) {
    struct dpvtOut *dpvt;
    struct plcInfo *pPlc;
    struct plcMessage *pMsg;
    struct wrCache *pcache;
    long status;
    
    if (devDnMpfDebug > 0)
	printf ("devXoDnMpf: init_input invoked for \"%s\"\n", prec->name);
    
    dpvt = (struct dpvtOut *) (calloc(1, sizeof(struct dpvtOut)));
    if (dpvt == NULL) {
	errlogPrintf("devXoDnMpf: calloc failed for \"%s\"\n", prec->name);
	prec->pact = TRUE;
	return S_rec_outMem;
    }
    prec->dpvt = (void *) dpvt;
    pMsg = &dpvt->msg;
    
    status = dnMpfAddr(prec, &dpvt->plcAddr, plink);
    if (status) {
	prec->pact = TRUE;
	return status;
    }
    pPlc = dpvt->plcAddr.plcInfo;
    
    dpvt->precord = prec;
    dpvt->type    = type;
    
    pMsg->server   = pPlc->server;
    pMsg->location = pPlc->location;
    pMsg->callback = devXoDnCallback;
    
    status = initDnMpfClient(pMsg);
    if (status) {
	prec->pact = TRUE;
	return status;
    }
    
    if ((dpvt->plcAddr.vAddr < WRITEMINADDR+DNREFOFFSET) ||
	(dpvt->plcAddr.vAddr > WRITEMAXADDR+DNREFOFFSET)) {
	errlogPrintf("devXoDnMpf: PLC address write-protected for \"%s\"\n",
		     prec->name);
	prec->pact = TRUE;
	return S_dev_badSignal;
    }
    
    pcache = pPlc->wrCache;
    if (pcache == NULL) {
	/* Not defined?  Create it */
	pcache = (struct wrCache *) calloc(1, sizeof (struct wrCache));
	if (pcache == NULL) {
	    errlogPrintf("devXoDnMpf: calloc failed for \"%s\"\n", prec->name);
	    prec->pact = TRUE;
	    return S_rec_outMem;
	}
	pPlc->wrCache = pcache;
    }
    
    dpvt->wrItem = &pcache->item[dpvt->plcAddr.vAddr - WRITEMINADDR+DNREFOFFSET];
    
    pMsg->cmd   = (pPlc->slaveId << 8) | WRITEVMEM;
    pMsg->len   = DN_PLCWORDLEN;
    pMsg->addr  = dpvt->plcAddr.vAddr;
    pMsg->pdata = dpvt->wrData;
    
    return 0;
}

static long init_ao(struct aoRecord *prec) {
    long status;
    
    if (devDnMpfDebug > 0)
	printf ("devXoDnMpf: Init ao invoked\n");
    
    status = init_output((struct dbCommon *) prec, AO, &prec->out);
    if (status)
	prec->pact = TRUE;	/* Error, prevent processing */
    
    return status;
}


static long init_bo(struct boRecord *prec) {
    long status;

    if (devDnMpfDebug > 0)
	printf ("devXoDnMpf: Init bo invoked\n");
 
    status = init_output((struct dbCommon *) prec, BO, &prec->out);
    if (status) {
	prec->pact = TRUE;	/* Error, prevent processing */
	return status;
    }
    
    prec->mask = 1 << ((struct dpvtOut *) prec->dpvt)->plcAddr.bitNum;
    return status;
}


static long init_mbbo(struct mbboRecord *prec) {
    long status;
    
    if (devDnMpfDebug > 0)
       printf ("devXoDnMpf: Init mbbo invoked\n");
 
    status = init_output((struct dbCommon *) prec, MBBO, &prec->out);
    if (status) {
	prec->pact = TRUE;	/* Error, prevent processing */
	return status;
    }
    
    prec->shft = ((struct dpvtOut *) prec->dpvt)->plcAddr.bitNum;
    prec->mask <<= prec->shft;
    return status;
}


static long init_mbbod(struct mbboDirectRecord *prec) {
    long status;
    
    if (devDnMpfDebug > 0)
       printf ("devXoDnMpf: Init mbbo invoked\n");
 
    status = init_output((struct dbCommon *) prec, MBBOD, &prec->out);
    if (status) {
	prec->pact = TRUE;	/* Error, prevent processing */
	return status;
    }
    
    prec->shft = ((struct dpvtOut *) prec->dpvt)->plcAddr.bitNum;
    prec->mask <<= prec->shft;
    return status;
}


static void setup_write(struct dbCommon *prec) {
    struct dpvtOut *dpvt = (struct dpvtOut *) prec->dpvt;
    struct wrItem *pitem = dpvt->wrItem;
    unsigned long mask;
 
    if (devDnMpfDebug >= 35)
        printf ("devXoDnMpf: setup_write entered for \"%s\"\n", prec->name);

    taskLock();
    switch(dpvt->type) {
	struct aoRecord *ao;
	struct boRecord *bo;
	struct mbboRecord *mbbo;
	struct mbboDirectRecord *mbbod;
	
	case AO:
	    ao = (struct aoRecord *) prec;
	    pitem->word = ao->rval;
	    break;
	
	case BO:
	    bo = (struct boRecord *) prec;
	    mask = bo->mask;
	    pitem->word = (pitem->word & ~mask) | (bo->rval & mask);
	    break;
	
	case MBBO:
	    mbbo = (struct mbboRecord *) prec;
	    mask = mbbo->mask;
	    pitem->word = (pitem->word & ~mask) | (mbbo->rval & mask);
	    break;
	
	case MBBOD:
	    mbbod = (struct mbboDirectRecord *) prec;
	    mask = mbbod->mask;
	    pitem->word = (pitem->word & ~mask) | (mbbod->rval & mask);
	    break;
    }
    dpvt->wrData[0] = pitem->word & 0xff;
    dpvt->wrData[1] = (pitem->word >> 8) & 0xff;
    taskUnlock();

    if (devDnMpfDebug >= 10) {
	printf("devXoDnMpf: Send V%o = %d\n", dpvt->msg.addr, *dpvt->msg.pdata);
    }
}


static long write_data(struct dbCommon *prec) {
    struct dpvtOut *dpvt=(struct dpvtOut *)prec->dpvt;
    struct plcMessage *pMsg;
    struct plcInfo *pPlc;

    if (devDnMpfDebug >= 35)
       printf ("devXoDnMpf: write_data called\n");

    if (!dpvt) return S_dev_NoInit;

    pMsg = &dpvt->msg;
    pPlc = dpvt->plcAddr.plcInfo;
    
    if (!prec->pact) {
	/* Record is idle, start a transaction */
	dpvt->tickStart = tickGet();
	
	if (devDnMpfDebug >= 10) 
	    printf("devXoDnMpf: Sending data from \"%s\"\n",
		   prec->name);
	
	/* Send the request */
	setup_write(prec);
	if (dnMpfClientSend(pMsg)) {
	    errlogPrintf("devXoDnMpf: MPF Send by \"%s\" failed\n", prec->name);
	    recGblSetSevr(prec, WRITE_ALARM, MAJOR_ALARM);
	    pPlc->nMpFail++;
	    return ERROR;
	}
	pPlc->nWrReqs++;
	prec->pact=TRUE;
    } else {
	/* Record busy, a transaction via MPF has completed */
	if (devDnMpfDebug >= 10) {
	    printf("devXoDnMpf: Got a reply for \"%s\"\n", 
		    prec->name);
	    printf("devXoDnMpf: Write V%o = %d returned status %d\n",
		    pMsg->addr, *pMsg->pdata, pMsg->status);
	} else if (pMsg->status && devDnMpfDebug >= 5)
	    printf("devXoDnMpf: Write reply for \"%s\" has status %d\n",
		   prec->name, pMsg->status);
	
	switch (pMsg->status) {
	case DN_SUCCESS:	/* OK reply was received */
	    pPlc->alarm = NO_ALARM;
	    pPlc->nSuccess++;

	    if (devDnMpfDebug >= 3)
		printf("devXoDnMpf: Write for \"%s\" took %lu ticks\n", 
			prec->name, tickGet() - dpvt->tickStart);
	    break;
	
	case DN_RETRY:
	case DN_TIMEOUT:
	case DN_RETRIED:
	case DN_BADARGS:
	case DN_ABORT:
	    errlogPrintf("devXoDnMpf: Comms error for \"%s\" status %d\n",
			prec->name, pMsg->status);
	    recGblSetSevr(prec, WRITE_ALARM, INVALID_ALARM); 
	    pPlc->alarm = INVALID_ALARM;
	    pPlc->nDnFail++;
	    break;
	    
	default:
	    errlogPrintf("devXiDnMpf: Error from MPF server \"%s\" status %d\n",
			 pMsg->server, pMsg->status);
	    recGblSetSevr(prec, WRITE_ALARM, MAJOR_ALARM);
	    pPlc->alarm = MAJOR_ALARM;
	    pPlc->nMpFail++;
	    return ERROR;
	}
    }

    return 0;
}

