/******************************************************************************

Project:
    DirectNet MPF

File:
    directNetServer.cc

Description:
    MPF Server for directNet serial protocol.

Author:
    Andrew Johnson
Created:
    August 2001
Version:
    directNetServer.cc,v 1.1 2002/01/07 22:36:20 anj Exp

Message Traffic:
    Server accepts only a Char8ArrayMessage command, where
	msg.cmd contains the following bit definitions:
		bits 8-15 are the directNet target slave ID,
		bit  7 contains 1 for write or 0 for read,
		bits 4-6 must be 0,
		bits 0-3 are the directNet data type
	msg.address holds the PCL start address
	msg.extra is the number of bytes to read or write
	msg.value contains any data to write (getSize() == msg.extra)
    The reply to this is another Char8ArrayMessage, where
	msg.cmd, msg.address and msg.extra are copied from the command
	msg.status is DN_SUCCESS or some other DN_ error value
	msg.value contains any data read

******************************************************************************/

#include <vxWorks.h>
#include <stdio.h>
#include <taskLib.h>
#include <string.h>

#include "Char8ArrayMessage.h"
#include "SerialPort.h"

#include "directNetMpf.h"


#define DNSERVER_TASKPRIO	100
#define DNSERVER_TASKSTACK	2000


int dnServerDebug = 0;


class dnServer {
public: // Published Interface
    dnServer(const char *name, const char *portName, int queueSize);
    ~dnServer();
    bool initOk() const;
    void stopServer();
    static STATUS doServer(int server);
    
private: // Internal routines
    void run();
    byteHandlerRC input(unsigned char data);
    
    static byteHandlerRC byteHandler(void *server, unsigned char data);
    
    void dnEOT();
    int32 dnSend(int len, unsigned char *pdata, double timeout);
    unsigned char dnLRC(int len, unsigned char *pdata);
    int32 dnSelect(int target);
    int32 dnHeader(uint32 cmd, int addr, int len);
    int32 dnBlock(int len, const char *pdata);
    int32 dnWrite(uint32 cmd, int addr, int len, const char *pdata);
    int32 dnRead(uint32 cmd, int addr, int len, char *pdata);
    
private: // Data
    const char *serverName;
    SerialPort *pPort;
    MessageServer *pMsgServer;
    bool stop;
    
    // The byteHandler uses a state machine:
    enum bhState {	// Comments indicate a state that ends a read
	idle,
	abort,		// Send EOT
	gotSeqChar,
	gotSlaveId,
	selectedOk,	// ENQ: SEQ, Target Addr, ENQ chars
	sentOk, 	// WRDATA: ACK
	retry,		// WRDATA: NAK
	dataOk,		// RDDATA: STX, Data block, ETB, checksum
	dataBad,	// RDDATA: wrong ETX/ETB/checksum
	dataChars,
	gotData,
	gotEtbChar,
	gotEtxChar,
	endDataOk	// RDDATA: STX, Last Data block, ETX, checksum
    };
    
    // Communication from byteHandler to task:
    bhState state;
    const char *error;
    
    // Communication from task to byteHandler:
    unsigned char slaveId;
    unsigned char *pReadData;
    int dataLen;
    
    // Internal use by byteHandler:
    int blockLen;
    unsigned char lrc;
    unsigned char *pBlock;
};


extern "C" {

STATUS startDirectNetServer(const char *name, const char *portName, int qSize)
{
    dnServer *pserver = new dnServer(name, portName, qSize);
    if (!pserver->initOk()) {
	puts("startDirectNetServer: Server creation failed");
	delete pserver;
	return ERROR;
    }
    
    if (dnServerDebug)
	printf("startDirectNetServer: dnServer \"%s\" object is at %p\n",
		name, pserver);
    int taskId = taskSpawn(const_cast<char*>(name), DNSERVER_TASKPRIO,
			   VX_FP_TASK, DNSERVER_TASKSTACK,
			   (FUNCPTR) dnServer::doServer, int(pserver),
			   0, 0, 0, 0, 0, 0, 0, 0, 0);
    if (taskId == ERROR) {
	perror("startDirectNetServer");
	return ERROR;
    }
    if (dnServerDebug)
	printf("startDirectNetServer: Task \"%s\" id = %#x\n",
		name, taskId);
    return OK;
}

} /* extern "C" */


STATUS dnServer::doServer(int server)
{
    dnServer *pserver = (dnServer *) server;
    pserver->run();
    return OK;
}


byteHandlerRC dnServer::byteHandler(void *server, unsigned char data)
{
    dnServer *pserver = (dnServer *) server;
    return pserver->input(data);
}


dnServer::dnServer(const char *name, const char *portName, int qSize) :
    serverName(name),
    pPort(SerialPort::bind(portName, this, dnServer::byteHandler)),
    pMsgServer(new MessageServer(name, qSize)),
    stop(false),
    state(idle)
{}

dnServer::~dnServer()
{
    // delete pMsgServer;
    if (pPort)
	pPort->release();
}

bool dnServer::initOk() const
{
    return (pPort && pMsgServer);
}

void dnServer::stopServer()
{
    stop = true;
}

void dnServer::run()
{
    while (!stop) {
	pMsgServer->waitForMessage();
	Message *pmsg;
	
	while (!stop && (pmsg = pMsgServer->receive())) {
	    if (dnServerDebug > 5)
		printf("dnServer %s: Received msg %p\n", serverName, pmsg);

	    if (pmsg->getType() != messageTypeChar8Array) {
		printf("dnServer %s: bad message type %d\n", serverName, pmsg->getType());
		delete pmsg;
		continue;
	    }
	    
	    Char8ArrayMessage *pcmd = (Char8ArrayMessage *) pmsg;
	    Char8ArrayMessage *pret = (Char8ArrayMessage *)
		pMsgServer->allocReplyMessage(pcmd, messageTypeChar8Array);
	    
	    // Allow multi-message clients to match our reply to their command
	    pret->cmd = pcmd->cmd;
	    pret->address = pcmd->address;
	    pret->extra = pcmd->extra;
	    
	    if (pcmd->cmd & WRITECMD) {
		if (dnServerDebug > 10)
		    printf("dnServer %s: Write cmd=%#x, addr=%#o, extra=%d\n",
			    serverName, pcmd->cmd, pcmd->address, pcmd->extra);
		if (pcmd->extra != pcmd->getSize()) {
		    printf("dnServer %s: size mismatch in write cmd, %d != %d\n",
			   serverName, pcmd->extra, pcmd->getSize());
		    delete pmsg;
		    continue;
		}
		pret->status = dnWrite(pcmd->cmd, pcmd->address, pcmd->extra, pcmd->value);
	    } else {
		if (dnServerDebug > 10)
		    printf("dnServer %s: Read cmd=%#x, addr=%#o, extra=%d\n",
			    serverName, pcmd->cmd, pcmd->address, pcmd->extra);
		pret->allocValue(pcmd->extra);
		pret->status = dnRead(pcmd->cmd, pcmd->address, pcmd->extra, pret->value);
		if (pret->status == DN_SUCCESS)
		    pret->setSize(pcmd->extra);
	    }
	    if (dnServerDebug > 10)
		printf("dnServer %s: Returning status=%d\n",
			serverName, pret->status);
	    pMsgServer->reply(pret);
	    delete pmsg;
	}
    }
}


void dnServer::dnEOT()
{
    if (dnServerDebug > 20)
	printf("dnServer %s: dnEOT()\n", serverName);
    
    unsigned char data = EOTCHAR;
    serialPortStatus status = pPort->write(&data, 1, ACKEOTDELAY);
    if (status != serialPortSuccess)
	printf("dnServer %s: EOT write failed/timed out\n", serverName);
    
    state = idle;	// Reset state machine...
}

int32 dnServer::dnSend(int len, unsigned char *pdata, double timeout)
{
    if (dnServerDebug > 15)
	printf("dnServer %s: dnSend(%d, %p, %f)\n",
		serverName, len, pdata, timeout);
    
    timeout += double(len) / BYTERATE;	// BYTERATE should come from port
    
    serialPortStatus status = pPort->writeRead(pdata, len, timeout);
    switch (status) {
    case serialPortSuccess:
	if (dnServerDebug > 20)
	    printf("dnServer %s: send Ok, state %d\n", serverName, state);
	return DN_SUCCESS;
	
    case serialPortTimeout:
	if (dnServerDebug > 20)
	    printf("dnServer %s: send timed out, state %d\n", serverName, state);
	return DN_TIMEOUT;
	
    case serialPortFailure:
	if (state == abort)
	    printf("dnServer %s: %s\n", serverName, error);
	if (dnServerDebug > 20)
	    printf("dnServer %s: send failed, state %d\n", serverName, state);
	return DN_RETRY;
    }
    printf("dnServer %s: Unknown status %d from serialPort", serverName, status);
    return DN_INTERNAL;
}

unsigned char dnServer::dnLRC(int len, unsigned char *pdata)
{
    if (dnServerDebug > 15)
	printf("dnServer %s: dnLRC(%d, %p)\n", serverName, len, pdata);
    
    unsigned char lrc = 0;
    
    for (int i = 0; i < len; i++) {
    	lrc ^= pdata[i];
    }
    
    return lrc;
}

int32 dnServer::dnSelect(int target)
{
    if (dnServerDebug > 15)
	printf("dnServer %s: dnSelect(%d)\n", serverName, target);
    
    if ((state != idle) && (state != abort)) {
	printf("dnServer %s: Entered dnSelect with state=%d\n", serverName, state);
	return DN_BADSTATE;
    }
    
    slaveId = target + SLAVEOFFSET;
    
    unsigned char abtsel[4];
    abtsel[0] = EOTCHAR;
    abtsel[1] = SEQCHAR;
    abtsel[2] = slaveId;
    abtsel[3] = ENQCHAR;
    unsigned char *select = &abtsel[1];
    
    int retries = MAX_RETRIES;
    do {
	int32 status = dnSend(3, select, ENQACKDELAY);
	if (status != DN_SUCCESS)
	    return status;
	
	if (state == abort)
	    select = abtsel;		// cancel and try ENQ again
    } while ((state == abort) && (--retries > 0));
    return (state == selectedOk) ? DN_SUCCESS : DN_RETRIED;
}


int32 dnServer::dnHeader(uint32 cmd, int addr, int len)
{
    if (dnServerDebug > 15)
	printf("dnServer %s: dnHeader(%u, %d, %d)\n", serverName, cmd, addr, len);
    
    if((cmd  != (cmd  & 0xffff)) ||
       (addr != (addr & 0xffff)) ||
       (len  != (len  & 0xffff)))
	return DN_BADARGS;
    
    int32 status = dnSelect((cmd >> 8) & 0xff);
    if (status)
	return status;
    
    if (state != selectedOk) {
	printf("dnServer %s: Returned from dnSelect with state=%d\n", serverName, state);
    	return DN_BADSTATE;
    }
    
    unsigned char header[HEADER_LEN];
    sprintf((char *) header, "%c%4.4X%4.4X%4.4X%2.2X%c", SOHCHAR, 
	    cmd, addr, len, MASTERID, ETBCHAR);
    header[HEADER_LEN-1] = dnLRC(HEADER_LEN - 3, &header[1]);
    
    int retries = MAX_RETRIES;
    do {
	status = dnSend(HEADER_LEN, header, HDRACKDELAY);
	if (status != DN_SUCCESS)
	    return status;
    } while ((state == selectedOk) && (--retries > 0));
    return ((state == sentOk) || (state == dataOk) || (state == endDataOk)) ?
	    DN_SUCCESS : DN_RETRIED;
}

int32 dnServer::dnBlock(int len, const char *pdata)
{
    if (dnServerDebug > 15)
	printf("dnServer %s: dnBlock(%d, %p)\n", serverName, len, pdata);
    
    if (state != sentOk) {
	printf("dnServer %s: Entered dnBlock with state=%d\n", serverName, state);
    	return DN_BADSTATE;
    }
    
    int blen = len;
    if (blen > 256) blen = 256;
    len -= blen;
    
    unsigned char block[256+3];
    block[0] = STXCHAR;
    memcpy(&block[1], pdata, blen);
    block[blen+1] = len ? ETBCHAR : ETXCHAR;
    block[blen+2] = dnLRC(blen, &block[1]);
    
    int retries = MAX_RETRIES;
    do {
	int32 status = dnSend(blen+3, block, DATACKDELAY);
	if (status != DN_SUCCESS)
	    return status;
    } while ((state == retry) && (--retries > 0));
    return (state == sentOk) ? DN_SUCCESS : DN_RETRIED;
}

int32 dnServer::dnWrite(uint32 cmd, int addr, int len, const char *pdata)
{
    if (dnServerDebug > 15)
	printf("dnServer %s: dnWrite(%u, %d, %d, %p)\n",
		serverName, cmd, addr, len, pdata);
    
    pReadData = 0;		// Signal to byteHandler
    
    int32 status = dnHeader(cmd, addr, len);
    if (status == DN_SUCCESS) {
	do {
	    status = dnBlock(len, pdata);
	    len -= 256;
	    pdata += 256;
	} while ((status == DN_SUCCESS) && (len > 0));
    }
    
    // Tidy up
    dnEOT();
    return status;
}


int32 dnServer::dnRead(uint32 cmd, int addr, int len, char *pdata)
{
    if (dnServerDebug > 15)
	printf("dnServer %s: dnRead(%u, %d, %d, %p)\n",
		serverName, cmd, addr, len, pdata);
    
    dataLen = len;				// Prepare for response
    pReadData = (unsigned char *) pdata;	// before sending header
    
    int32 status= dnHeader(cmd, addr, len);
    if (status == DN_SUCCESS) {
	if ((state != dataOk) && (state != endDataOk)) {
	    dnEOT();
	    printf("dnServer %s: Returned from dnHeader with state=%d\n", serverName, state);
	    return DN_BADSTATE;
	}
	
	int retries = MAX_RETRIES;
	do {
	    unsigned char reply;
	    if ((state == dataOk) || (state == endDataOk)) {
		reply = ACKCHAR;
		retries = MAX_RETRIES;
	    } else if (state == dataBad) {
		reply = NAKCHAR;
		if (--retries <= 0) {
		    status = DN_RETRIED;
		    break;
		}
	    } else if (state == idle) {
		status = DN_SUCCESS;
		break;
	    } else if (state == abort) {
		status = DN_ABORT;
		break;
	    } else {
		status = DN_BADSTATE;
		break;
	    }
	    status = dnSend(1, &reply, DATACKDELAY);
	} while (status == DN_SUCCESS);
    }
    
    // Tidy up
    dnEOT();
    return status;
}


byteHandlerRC dnServer::input(unsigned char data)
{
    switch (state) {
    case idle:
    case abort:
	if (data == SEQCHAR)
	    state = gotSeqChar;
	break;
	
    case gotSeqChar:
	if (data == slaveId) {
	    state = gotSlaveId;
	} else {
	    state = abort;
	    error = "Bad target address in ENQ reply";
	    return byteHandlerError;
	}
	break;
	
    case gotSlaveId:
	if (data == ACKCHAR) {
	    state = selectedOk;
	    return byteHandlerEndRead;
	} else if (data == NAKCHAR) {
	    state = abort;
	    return byteHandlerEndRead; // retry the ENQ send
	} else {
	    state = abort;
	    error = "Bad ACK/NAK in ENQ reply";
	    return byteHandlerError;
	}
	break;
	
    case selectedOk:
	if (data == ACKCHAR) {
	    if (pReadData == 0) {	// Write command
		state = sentOk;
		return byteHandlerEndRead;
	    } else {			// Read command
		state = dataOk;		// Wait for first block too
	    }
	} else if (data == NAKCHAR) {
	    return byteHandlerEndRead;	// retry the header send
	} else {
	    state = abort;
	    error = "Bad ACK/NAK in HDR reply";
	    return byteHandlerError;
	}
	break;
	
    case sentOk:
    case retry:
	if (data == ACKCHAR) {
	    state = sentOk;
	    return byteHandlerEndRead;	// block sent Ok
	} else if (data == NAKCHAR) {
	    state = retry;
	    return byteHandlerEndRead;	// retry block send
	} else {
	    state = abort;
	    error = "Bad ACK/NAK in DATA reply";
	    return byteHandlerError;
	}
	break;
	
    case dataOk:
    case dataBad:
	if (data == STXCHAR) {
	    state = dataChars;		// data block starting
	    if (dataLen > 256) {
		blockLen = 256;
	    } else {
		blockLen = dataLen;
	    }
	    pBlock = pReadData;		// save block start position
	    lrc = 0;
	} else if (data == EOTCHAR) {
	    state = abort;
	    error = "Early EOT from target";
	    return byteHandlerError;
	}
	// else keep looking for STX or EOT until timeout
	break;
	
    case dataChars:
	*pReadData++ = data;
	lrc ^= data;
	if (--blockLen <= 0) {
	    state = gotData;
	}
	break;
	
    case gotData:
	if (data == ETBCHAR) {
	    state = gotEtbChar;
	} else if (data == ETXCHAR) {
	    state = gotEtxChar;
	} else {
	    state = dataBad;		// send NAK
	    pReadData = pBlock;		// Reset to last block start
	    return byteHandlerEndRead;
	}
	break;
	
    case gotEtbChar:
	if ((data == lrc) && (dataLen > 256)) {
	    dataLen -= 256;
	    state = dataOk;		// send ACK
	} else {
	    state = dataBad;		// send NAK
	    pReadData = pBlock;		// Reset to last block start
	}
	return byteHandlerEndRead;
	
    case gotEtxChar:
	if (data == lrc) {
	    if (dataLen > 256) {	// was expecting more
		state = abort;
		error = "Short read data reply";
		return byteHandlerError;
	    }
	    state = endDataOk;		// send ACK
	} else {
	    state = dataBad;		// send NAK
	    pReadData = pBlock;		// Reset to last block start
	}
	return byteHandlerEndRead;
	
    case endDataOk:
	if (data == EOTCHAR) {
	    state = idle;
	    return byteHandlerEndRead;
	}
	state = abort;
	error = "Bad EOT after read";
	return byteHandlerError;
    }
    return byteHandlerOK;
}

